import { singleton } from 'zutils'
import Client from 'ioredis'
import Redlock, { Lock, ResourceLockedError } from 'redlock'
import logger from 'logger/logger'

interface IRequest {
  method: string
  url: string
  user?: {
    id: string
  }
  lock?: Lock
}

const redisA = new Client(process.env.REDIS)
const redlock = new Redlock(
  // You should have one client for each independent redis node
  // or cluster.
  [redisA],
  {
    // The expected clock drift; for more details see:
    // http://redis.io/topics/distlock
    driftFactor: 0.01, // multiplied by lock ttl to determine drift time

    // The max number of times Redlock will attempt to lock a resource
    // before erroring.
    retryCount: 0,

    // the time in ms between attempts
    retryDelay: 200, // time in ms

    // the max time in ms randomly added to retries
    // to improve performance under high contention
    // see https://www.awsarchitectureblog.com/2015/03/backoff.html
    retryJitter: 200, // time in ms

    // The minimum remaining time on a lock before an extension is automatically
    // attempted with the `using` API.
    automaticExtensionThreshold: 500, // time in ms
  },
)

redlock.on('error', error => {
  // Ignore cases where a resource is explicitly marked as locked on a client.
  if (error instanceof ResourceLockedError) {
    return
  }
  // Log all other errors.
  logger.error(error)
})

@singleton
export class SyncLocker {
  public async unlock(req: IRequest) {
    if (req.lock) {
      await req.lock.release()
    }
  }

  public async checkLock(req: IRequest) {
    const key = `${req.method}:${req.url}:${req.user?.id || ''}`
    let lock = await redlock.acquire([key], 3000)
    req.lock = lock
    return true
  }
}
