import { MANUAL_OPEN_GAME, SCORE_GAME_STEP } from 'common/Constants'
import { ActivityGame } from 'models/ActivityGame'
import { DAILY_SIGN, SIGN_SEQ, SIGN_TOTAL, TicketRecord, USE_TICKET } from 'models/TicketRecord'
import { queryCheckInList } from 'services/chain.svr'
import { checkInToday, seqSignCfg, seqSignScore, totalSignCfg, totalSignScore } from 'services/sign.svr'
import { ZError, BaseController, router } from 'zutils'
import { SyncLocker } from 'common/SyncLocker'
import { formatDate } from 'utils/utcdate.util'
import { generateStepReward } from 'services/game.svr'
import { updateRankScore } from 'services/rank.svr'
import { ExploreRecord } from 'models/ExploreRecord'
import { isObjectId } from 'zutils/utils/string.util'
import { GeneralScription } from 'models/chain/GeneralScription'
import { CheckIn } from 'models/chain/CheckIn'
import { formatNumShow, isValidDayTag } from 'common/Utils'
import logger from 'logger/logger'

/**
 * 探索游戏相关接口
 */
class GameController extends BaseController {
  /**
   * 签到
   */
  @router('post /api/user/checkin')
  async checkIn(req) {
    await new SyncLocker().checkLock(req)
    logger.db('checkin', req)
    const user = req.user
    const { address } = user
    let { day, days } = req.params
    day = day || days
    if (day) {
      if (!isValidDayTag(day)) {
        throw new ZError(11, 'invalid day')
      }
    }
    const dateTag = day || formatDate(new Date())
    const exists = await TicketRecord.find({ user: user.id, activity: user.activity, type: DAILY_SIGN })
    const existsSet = new Set()
    exists.forEach(record => {
      // @ts-ignore
      let _dateTag = record.data?.dataTag || formatDate(record.createdAt)
      existsSet.add(_dateTag)
    })
    if (existsSet.has(dateTag)) {
      throw new ZError(12, 'already claimed')
    }
    const gameRecord = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && gameRecord.status === 0) {
      throw new ZError(11, 'map not open')
    }
    const record = await checkInToday(address, dateTag)
    if (!record) {
      throw new ZError(13, 'had not check in')
    }
    const ticketRecord = new TicketRecord({
      user: user.id,
      activity: user.activity,
      type: DAILY_SIGN,
      data: { dataTag: dateTag },
      score: 1,
    })
    await ActivityGame.updateOne({ user: user.id, activity: user.activity }, { $inc: { tickets: 1 } })
    await ticketRecord.save()
    return { ticket: 1 }
  }

  /**
   * 签到列表
   */
  @router('get /api/user/checkin/list/:tag')
  async checkInList(req) {
    const user = req.user
    const { tag } = req.params
    let days: any = 1
    if (tag === '1month') {
      days = '1month'
    } else if (tag === 'last') {
      days = '1'
    } else if (tag === 'all') {
      const start = new Date(req.activity.start).getTime()
      // calc days from start to now
      const now = new Date().getTime()
      days = Math.floor((now - start) / 86400000)
    }
    const res = await queryCheckInList(user.address.toLowerCase(), days, 0)
    const ticketRecords = await TicketRecord.find({ user: user.id, activity: user.activity, type: DAILY_SIGN })
    const claimedSet = new Set()
    ticketRecords.forEach(record => {
      // @ts-ignore
      let dateTag = record.data?.dataTag || formatDate(record.createdAt)
      claimedSet.add(dateTag)
    })
    for (let record of res) {
      record.claimed = claimedSet.has(record.day)
    }
    return res
  }
  /**
   * 领取累计签到奖励
   */
  @router('post /api/user/checkin/claim')
  async claimCheckResult(req) {
    await new SyncLocker().checkLock(req)
    logger.db('claim_checkin_total', req)
    const user = req.user
    let { days } = req.params
    if (!days || isNaN(days)) {
      throw new ZError(11, 'invalid days')
    }
    days = parseInt(days)
    if (days < 1) {
      throw new ZError(12, 'invalid days')
    }
    const checkRecord = await CheckIn.findOne({ from: user.address.toLowerCase() }).sort({ _id: -1 })
    if (!checkRecord) {
      throw new ZError(12, 'no onchain check in record')
    }
    if (days > checkRecord.total) {
      throw new ZError(13, 'invalid days')
    }
    const ticketRecords = await TicketRecord.find({ user: user.id, activity: user.activity, type: SIGN_TOTAL })
    const claimedSet = new Set()
    ticketRecords.forEach(record => {
      claimedSet.add(record.data.day)
    })
    if (claimedSet.has(days)) {
      throw new ZError(14, 'already claimed')
    }

    const score = totalSignScore(days)
    if (score === 0) {
      throw new ZError(15, 'invalid days')
    }
    const ticketRecord = new TicketRecord({
      user: user.id,
      activity: user.activity,
      type: SIGN_TOTAL,
      data: { day: days },
      score,
    })
    const gameRecord = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && gameRecord.status === 0) {
      throw new ZError(12, 'map not open')
    }
    await ActivityGame.updateOne({ user: user.id, activity: user.activity }, { $inc: { tickets: score } })
    await ticketRecord.save()
    return { ticket: score }
  }
  /**
   * 领取连续签到奖励
   */
  @router('post /api/user/checkin/claim_seq')
  async claimCheckSeqResult(req) {
    await new SyncLocker().checkLock(req)
    logger.db('claim_checkin_seq', req)
    const user = req.user
    let { days } = req.params
    if (!days || isNaN(days)) {
      throw new ZError(11, 'invalid days')
    }
    days = parseInt(days)
    if (days < 1) {
      throw new ZError(12, 'invalid days')
    }
    const checkRecord = await CheckIn.findOne({ from: user.address.toLowerCase() }).sort({ _id: -1 })
    if (!checkRecord) {
      throw new ZError(12, 'no onchain check in record')
    }
    if (days > checkRecord.total) {
      throw new ZError(13, 'invalid days')
    }
    const ticketRecords = await TicketRecord.find({ user: user.id, activity: user.activity, type: SIGN_SEQ })
    const claimedSet = new Set()
    ticketRecords.forEach(record => {
      claimedSet.add(record.data.day)
    })
    if (claimedSet.has(days)) {
      throw new ZError(14, 'already claimed')
    }

    const score = seqSignScore(days)
    if (score === 0) {
      throw new ZError(15, 'invalid days')
    }
    const ticketRecord = new TicketRecord({
      user: user.id,
      activity: user.activity,
      type: SIGN_SEQ,
      data: { day: days },
      score,
    })
    const gameRecord = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && gameRecord.status === 0) {
      throw new ZError(12, 'map not open')
    }
    await ActivityGame.updateOne({ user: user.id, activity: user.activity }, { $inc: { tickets: score } })
    await ticketRecord.save()
    return { ticket: score }
  }
  /**
   * 开启地图
   */
  @router('get /api/game/open')
  async openMap(req) {
    const user = req.user
    const gameInfo = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (gameInfo.status === 1) {
      throw new ZError(11, 'already opened')
    }
    // check if all task finished
    if (!user.allTaskFinished()) {
      throw new ZError(12, 'The opening conditions are not met')
    }
    gameInfo.status = 1
    await gameInfo.save()
    return {}
  }
  /**
   * 探索状态
   */
  @router('get /api/game/stat')
  async gameStat(req, res) {
    const user = req.user
    const record = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && record.status === 0) {
      throw new ZError(12, 'map not open')
    }
    const signCfg = seqSignCfg
    const dateTag = formatDate(new Date())
    const checkRecord = await checkInToday(user.address.toLowerCase(), dateTag)
    let todayStat = 0
    if (checkRecord) {
      // 有签到的链上记录, 说明已签到
      todayStat = 1
    }
    const claimedDaySet = new Set()

    const ticketRecords = await TicketRecord.find({
      user: user.id,
      activity: user.activity,
      type: { $in: [SIGN_TOTAL, USE_TICKET, SIGN_SEQ, DAILY_SIGN] },
    })
    const claimedSet = new Set()
    const seqSet = new Set()
    let totalUsed = 0
    ticketRecords.forEach(record => {
      if (record.type === USE_TICKET) {
        totalUsed += record.score
      } else if (record.type === SIGN_TOTAL) {
        claimedSet.add(record.data.day)
      } else if (record.type === SIGN_SEQ) {
        seqSet.add(record.data.day)
      } else if (record.type === DAILY_SIGN) {
        //@ts-ignore
        let dateTag = record.data?.dataTag || formatDate(record.createdAt)
        claimedDaySet.add(dateTag)
      }
    })
    if (claimedDaySet.has(dateTag)) {
      // 检查是否已领取
      todayStat = 9
    }
    const totalStat = []
    const lastRecord = await CheckIn.findOne({ from: user.address.toLowerCase() }).sort({ _id: -1 })
    for (let cfg of totalSignCfg) {
      let state = 0
      if (cfg.days <= lastRecord?.total || 0) {
        state = 1
      }
      if (claimedSet.has(cfg.days)) {
        state = 9
      }
      totalStat.push({
        days: cfg.days,
        tickets: cfg.reward,
        state,
      })
    }
    // 连续签到, 根据最大连续签到天数来判断是否拥有领取的资格
    const seqStat = []
    for (let cfg of seqSignCfg) {
      let state = 0
      if (cfg.days <= lastRecord?.maxSeq || 0) {
        state = 1
      }
      if (seqSet.has(cfg.days)) {
        state = 9
      }
      seqStat.push({
        days: cfg.days,
        tickets: cfg.reward,
        state,
      })
    }
    totalStat.sort((a, b) => a.days - b.days)
    return {
      ticket: record.tickets,
      totalUsed: Math.abs(totalUsed),
      signCfg,
      todayStat,
      todayTickets: 1,
      daysTotal: lastRecord?.total || 0,
      daysSeq: lastRecord?.count || 0,
      seqStat,
      totalStat,
    }
  }
  @router('post /api/game/pre_step')
  async gameChest(req) {
    const user = req.user
    await new SyncLocker().checkLock(req)
    let { step } = req.params
    step = step || '1'
    if (isNaN(step)) {
      throw new ZError(11, 'invalid step')
    }

    // check if step is safe int
    step = parseInt(step)
    if (step < 1) {
      step = 1
    }
    const record = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && record.status === 0) {
      throw new ZError(12, 'map not open')
    }
    if (record.tickets < step) {
      throw new ZError(13, 'insufficient tickets')
    }
    const exploreRecord = new ExploreRecord({
      user: user.id,
      activity: user.activity,
      step,
    })
    await exploreRecord.save()

    return { id: exploreRecord.id }
  }
  /**
   * 探索
   */
  @router('post /api/game/step')
  async gameStep(req, res) {
    await new SyncLocker().checkLock(req)
    logger.db('step', req)
    const user = req.user
    const { id } = req.params
    if (!id) {
      throw new ZError(11, 'invalid id')
    }
    if (!isObjectId(id)) {
      throw new ZError(12, 'invalid id')
    }
    if (process.env.NODE_ENV === 'production') {
      const chainRecord = await GeneralScription.findOne({
        from: user.address.toLowerCase(),
        op: 'explore',
        data: id,
      })
      if (!chainRecord) {
        throw new ZError(13, 'waiting for chain confirm')
      }
    }
    const exploreRecord = await ExploreRecord.findById(id)
    if (!exploreRecord) {
      throw new ZError(14, 'invalid id')
    }
    if (exploreRecord.status !== 0) {
      throw new ZError(15, 'invalid status')
    }
    const step = exploreRecord.step
    const record = await ActivityGame.insertOrUpdate({ user: user.id, activity: user.activity }, {})
    if (MANUAL_OPEN_GAME && record.status === 0) {
      throw new ZError(16, 'map not open')
    }
    if (record.tickets < step) {
      exploreRecord.status = -1
      await exploreRecord.save()
      throw new ZError(17, 'insufficient tickets')
    }
    exploreRecord.status = 1
    await exploreRecord.save()
    const ticketRecord = new TicketRecord({
      user: user.id,
      activity: user.activity,
      type: USE_TICKET,
      data: {},
      score: -step,
    })
    let updateData: any = {
      $inc: { tickets: -step },
    }
    let { score, chests } = generateStepReward(user.id, user.activity, step)
    // if (chests.length > 0) {
    //   updateData.maxNoChestCount = 0
    // } else {
    //   if (record.maxNoChestCount + step >= RESET_STEP) {
    //     updateData.maxNoChestCount = 0
    //     const level = generateChestLevel()
    //     chests.push(generateNewChest(user.id, user.activity, level, ChestStatusEnum.NORMAL))
    //   } else {
    //     updateData['$inc']['maxNoChestCount'] = step
    //   }
    // }
    let saveRes = await ActivityGame.findOneAndUpdate(
      { user: user.id, activity: user.activity, tickets: { $gte: step } },
      updateData,
      {
        new: false,
        upsert: false,
        includeResultMetadata: true,
      },
    )
    if (!saveRes.lastErrorObject.updatedExisting) {
      throw new ZError(17, 'insufficient tickets.')
    }
    await updateRankScore({
      user: user.id,
      score: score,
      activity: user.activity,
      scoreType: SCORE_GAME_STEP,
      scoreParams: {
        step: step,
      },
    })
    // batch save chests
    for (let chest of chests) {
      await chest.save()
    }
    await ticketRecord.save()

    return { score: formatNumShow(score), chests: chests.map(chest => chest.toJson()) }
  }
}
