import { DocumentType } from '@typegoose/typegoose'
import { AuthRecordClass } from 'models/oauth/AuthRecord'
import { timeoutFetch } from 'zutils/utils/net.util'
import { ZError } from 'zutils'
const DISCORD_API_HOST = 'https://discord.com/api/v10'
const DISCORD_APP_HOST = 'https://discordapp.com/api'
const DEFAULT_TIMEOUT = 30000

/**
 * 直接通过discord接口, 获取工会中某用户信息
 * 返回的数据结构:
 * https://discord.com/developers/docs/resources/guild#guild-member-object
 * //TODO: 须在.env中配置以下参数:
 * DISCROD_GUILD_ID
 * DISCORD_BOT_TOKEN
 */
export async function getGuildMember(guid: string, uid: string) {
  const url = `${DISCORD_API_HOST}/guilds/${guid}/members/${uid}`
  const token = process.env.DISCORD_BOT_TOKEN
  const response = await timeoutFetch(
    url,
    {
      headers: {
        authorization: `Bot ${token}`,
      },
    },
    DEFAULT_TIMEOUT,
  )
  const data = await response.json()
  return data
}

/**
 * 检查用户是否已经加入工会
 * @param uid
 * @returns
 */
export async function checkJoinGuld(guid: string, uid: string) {
  try {
    const res = await getGuildMember(guid, uid)
    /**
     * if user not in guild, res will be:
     * {
     *  "message": "Unknown User",
     *  "code": 10013
     *  }
     */
    if (res.code === 10013) {
      return false
    }
    if (res.code) {
      throw new Error(res.message)
    }
    return true
  } catch (e) {
    return false
  }
}
/**
 * 检查用户是否已经获得某个角色
 * @param uid
 * @param roleId
 * @returns
 */
export async function checkGotRole(guid: string, uid: string, roleId: string) {
  try {
    const res = await getGuildMember(guid, uid)
    if (res.code === 10013) {
      return false
    }
    if (res.code) {
      // 如果有其他错误, 直接抛出异常, 让客户端重试
      throw new Error(res.message)
    }
    if (res.roles.includes(roleId)) {
      return true
    }
    return false
  } catch (e) {
    return false
  }
}

export async function getAvableAccessToken(record: DocumentType<AuthRecordClass>) {
  if (record.expiresIn < Date.now() / 1000 + 60 * 5) {
    console.log('discord access token expired')
    const data = await refreshDiscordToken(record.refreshToken)
    if (data.code) {
      throw new ZError(30, 'refresh token error')
    }
    record.accessToken = data.access_token
    record.refreshToken = data.refresh_token
    record.expiresIn = (Date.now() / 1000 + data.expires_in || 0) | 0
    await record.save()
  }
  return record.accessToken
}
export async function refreshDiscordToken(rtoken: string) {
  const clientId = process.env.DISCORD_CLIENT_ID
  const clientSecret = process.env.DISCORD_CLIENT_SECRET
  const response = await timeoutFetch(
    'https://discord.com/api/oauth2/token',
    {
      method: 'POST',
      headers: {
        'Content-Type': 'application/x-www-form-urlencoded',
      },
      body: new URLSearchParams({
        grant_type: 'refresh_token',
        client_id: clientId,
        client_secret: clientSecret,
        refresh_token: rtoken,
      }),
    },
    DEFAULT_TIMEOUT,
  )
  const data = await response.json()
  return data
}

/**
 * 用户在某个工会的信息, oauth2方式
 * @param token
 * @param guildId
 * @returns
 * 返回的数据结构:
 * https://discord.com/developers/docs/resources/guild#guild-member-object
 */
export async function userGuildMember(token: string, guildId: string) {
  const url = `${DISCORD_APP_HOST}/users/@me/guilds/${guildId}/member`
  const response = await timeoutFetch(
    url,
    {
      headers: {
        authorization: `Bearer ${token}`,
      },
    },
    DEFAULT_TIMEOUT,
  )
  const data = await response.json()
  return data
}
