import { hmacSha256 } from 'zutils/utils/security.util'
import { handleFetch, timeoutFetch } from 'zutils/utils/net.util'
import { AuthRecord, PlatEnum } from 'models/oauth/AuthRecord'
import { ZError, ZRedisClient } from 'zutils'
import { getAvableAccessToken, userGuildMember } from './discord.svr'
import { logger } from 'ethers'

const SECRET_KEY = process.env.HASH_SALT
const DEFAULT_TIMEOUT = 30000

function createSign(address: string) {
  address = address.toLowerCase()
  let signCheck = hmacSha256(`address=${address}`, SECRET_KEY)
  return signCheck
}

export async function checkTwitter(address: string) {
  let record = await AuthRecord.findOne({ address: address.toLowerCase(), platform: PlatEnum.TWITTER })
  let result: any = {}
  let errcode = 0
  let errmsg = ''
  if (!record) {
    errcode = 20
    errmsg = 'not bind'
  } else {
    result.username = record.username
    result.userid = record.openId
    result.avatar = record.avatar
  }
  return { data: result, errcode, errmsg }
}

export async function checkDiscord(address: string) {
  let record = await AuthRecord.findOne({ address: address.toLowerCase(), platform: PlatEnum.DISCORD })
  let result: any = {}
  let errcode = 0
  let errmsg = ''
  if (!record) {
    errcode = 20
    errmsg = 'not bind'
  } else {
    result.verified = record.condition
    result.username = record.username
    result.userid = record.openId
    result.discriminator = record.discriminator
  }
  return { data: result, errcode, errmsg }
}

export async function checkDiscordGuid(address: string, gid: string) {
  let sign = createSign(address)
  const data = {
    gid,
    address,
    sign,
  }
  const url = `${process.env.OAUTH_SVR_URL}/activity/discord/guild`
  const options: any = {
    method: 'POST',
    headers: {
      'Content-Type': 'application/json; charset=utf-8',
    },
    body: JSON.stringify(data),
  }
  return timeoutFetch(url, options, DEFAULT_TIMEOUT).then(res => res.json())
}

export async function checkDiscordRole(address: string, gid: string, rid: string) {
  let record = await AuthRecord.findOne({ address, platform: PlatEnum.DISCORD })
  if (!record) {
    throw new ZError(20, 'need connect discord first')
  }
  const key = `oauth:${gid}:${address}`
  const cache = await new ZRedisClient().get(key)
  let roleSet = new Set()
  if (cache) {
    try {
      let roleArr = JSON.parse(cache)
      roleSet = new Set(roleArr)
      if (roleSet.has(rid)) {
        return { data: { result: true }, errcode: 0, errmsg: '' }
      }
    } catch (e) {
      console.error(e)
    }
  }
  try {
    let accessToken = await getAvableAccessToken(record)
    let data = await userGuildMember(accessToken, gid)
    roleSet = new Set(data.roles)
    let result = roleSet.has(rid)
    new ZRedisClient().pub.set(key, JSON.stringify([...roleSet]), 'EX', 300, () => {
      console.log('cache set success: ', key)
    })
    return { data: { result }, errcode: 0, errmsg: '' }
  } catch (err) {
    logger.info('checkDiscordRole', err?.message || err)
    return { data: { result: false }, errcode: 14, errmsg: 'had no role' }
  }
}
